//////////////////////////////////////////////////////////////////////////////////
// University of Toronto
// NES APU UNIT
// Engineer: Cedomir Segulja
// Create Date: 03/21/2008 
// Design Name: NES APU UNIT
//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////
// Top Level Module
//////////////////////////////////////////////////////////////////////////////////

////////////////////////////APU OPB REGISTER MAP//////////////////////////////////
//
//
// Rectangle 1 :
//  00 0 00 Register1 BASEADDR
//  01 1 04 Register2 BASEADDR + 4
//  02 2 08 Register3 BASEADDR + 8
//  03 3 0C Register4 BASEADDR + 12
//
//
// Rectangle 2 :
//  04 4 10 Register1 BASEADDR + 16
//  05 5 14 Register2 BASEADDR + 20
//  06 6 18 Register3 BASEADDR + 24
//  07 7 1C Register4 BASEADDR + 28
//
//
// Triangle :
//  08 8 20 Register1 BASEADDR + 32
//  09 9 24 NOT USED
//  0A 10 28 Register3 BASEADDR + 40
//  0B 11 2C Register4 BASEADDR + 44
//
//
// Noise :
//  0C 12  30 Register1 BASEADDR + 48
//  0D 13  34 NOT USED
//  0E 14  38 Register3 BASEADDR + 56
//  0F 15  3C Register4 BASEADDR + 60
//
//
// DMC :
//  10 16  40 Register1 BASEADDR + 64
//  11 17  44 Register2 BASEADDR + 68
//  12 18  48 Register3 BASEADDR + 72
//  13 19  4C Register4 BASEADDR + 76
//
//  14 20 50 NOT USED  
//  15 21 54 Status Register BASEADDR + 84
//  16 22 58 NOT USED
//  17 23 5C Frames BASEADDR + 92
//   
//////////////////////////////////////////////////////////////////////////////////

// uncomment this line for simulation
//`define SIMULATION 1

module apu_opb(
               iNES_clk,
               iOPB_clk,
               iOPB_rst,
               iOPB_select,
               iOPB_RNW,
               iOPB_ABus,
               iOPB_DBus,
               iOPB_MnGrant,
               iOPB_xferAck,
               oMn_request,
               oMn_ABus,
               oMn_DBus,
               oMn_Select,
               oMn_DBusEn,
               oMn_BE,
               oMn_RNW,
               oSln_DBus,
               oSln_xferAck,
               iAC97_Bit_Clk,
					iAC97_SData_In,
               oAC97_Sync,
               oAC97_SData_Out,
               oAC97_Reset
               );
					
  //-------------------------------------------------------------------------------------------------------
  //  PARAMETERS 
  //-------------------------------------------------------------------------------------------------------
  
  parameter C_BASEADDR = 32'hffffff00;
  parameter C_HIGHADDR = 32'hffffffff;
  parameter OPB_WIDTH = 32;
  parameter DMC_MEMORY = 32'h8040000;
					
  //-------------------------------------------------------------------------------------------------------
  //  PORTS
  //-------------------------------------------------------------------------------------------------------
	
  input iNES_clk;
  input iOPB_clk;
  input iOPB_rst;
  input iOPB_select;
  input iOPB_RNW;
  input [OPB_WIDTH-1:0] iOPB_ABus;
  input [OPB_WIDTH-1:0] iOPB_DBus;
  input iOPB_MnGrant;
  input iOPB_xferAck;
  output oMn_request;
  output [OPB_WIDTH-1:0] oMn_ABus;
  output [OPB_WIDTH-1:0] oMn_DBus;
  output oMn_Select;
  output oMn_DBusEn;
  output oMn_RNW;
  output [3:0] oMn_BE;
  output oSln_xferAck;
  output [OPB_WIDTH-1:0] oSln_DBus;
  input iAC97_Bit_Clk;
  input iAC97_SData_In;
  output oAC97_Sync;
  output oAC97_SData_Out;
  output oAC97_Reset;
     			  
  //-------------------------------------------------------------------------------------------------------
  //  AC 97 Codec Setup 
  //-------------------------------------------------------------------------------------------------------
  
  reg [1:0] AC97_state;
  
  parameter AC97_WAIT_FOR_CODEC_READY = 0;
  parameter AC97_SET_MASTER_VOLUME = 1;
  parameter AC97_SET_PCM_VOLUME = 2;
  parameter AC97_READY = 3;
   
  wire [6:0] AC97_Reg_Addr;
  wire AC97_Reg_Write_Strobe;
  wire AC97_Reg_Busy;
  wire AC97_codec_ready;
	 
  assign oAC97_Reset = !iOPB_rst;
  assign AC97_Reg_Write_Strobe = (AC97_state == AC97_SET_MASTER_VOLUME) || (AC97_state == AC97_SET_PCM_VOLUME);
  assign AC97_Reg_Addr = (AC97_state == AC97_SET_MASTER_VOLUME) ? 7'b0000010:7'b0011000;

  `ifdef SIMULATION
  always @ (posedge iOPB_clk or posedge iOPB_rst)
  `else
  always @ (posedge iAC97_Bit_Clk or posedge iOPB_rst)
  `endif
  begin
    if (iOPB_rst)
    begin
      AC97_state <= AC97_WAIT_FOR_CODEC_READY;
	 end else
    begin
      case (AC97_state)
		  AC97_WAIT_FOR_CODEC_READY:
		    if (AC97_codec_ready == 1)
			   AC97_state <= AC97_SET_MASTER_VOLUME;
		  AC97_SET_MASTER_VOLUME:
		    if (AC97_Reg_Busy == 0)
			   AC97_state <= AC97_SET_PCM_VOLUME;
		  AC97_SET_PCM_VOLUME:
		    if (AC97_Reg_Busy == 0)
			   AC97_state <= AC97_READY;  
		endcase
	 end	 
  end
   
  //-------------------------------------------------------------------------------------------------------
  //  OPB STATE MACHINE
  //-------------------------------------------------------------------------------------------------------
    
  reg [2:0] OPB_state; 
  
  parameter OPB_RESET_0 = 0;
  parameter OPB_RESET_1 = 1;
  parameter OPB_RESET_2 = 2;
  parameter OPB_READY = 3;
  parameter OPB_SLAVE_READ = 4;
  parameter OPB_SLAVE_WRITE = 5;
  parameter OPB_MASTER_REQUEST = 6;
  parameter OPB_MASTER_SELECT = 7;
  
  reg OPB_DMA_req;
  wire[OPB_WIDTH-1:0] DMA_addr;
  
  wire reset_channels; 
  wire ValidAddr;   
  
  assign reset_channels = OPB_state == OPB_RESET_0 || OPB_state == OPB_RESET_2;
  assign ValidAddr = (iOPB_ABus <= C_HIGHADDR) & (iOPB_ABus >= C_BASEADDR);
  
  always @(posedge iOPB_clk or posedge iOPB_rst)  
  begin
    if (iOPB_rst)
	   OPB_state <= OPB_RESET_0;
	 else
    begin
	   case (OPB_state)
	     OPB_RESET_0: 
		    OPB_state <= OPB_RESET_1;
	     OPB_RESET_1: 
		    OPB_state <= OPB_RESET_2;
	     OPB_RESET_2: 
		    if (AC97_state == AC97_READY) OPB_state <= OPB_READY;
	     OPB_READY: 
		  begin
		    //MASTER
			 if (OPB_DMA_req) OPB_state <= OPB_MASTER_REQUEST;	
			 else begin
			 //SLAVE
            if (iOPB_select & ValidAddr)
	         begin
		        if (iOPB_RNW)  
		          OPB_state <= OPB_SLAVE_READ;
		        else
		          OPB_state <= OPB_SLAVE_WRITE;
		      end
	       end
		  end
	     OPB_MASTER_REQUEST:
		    if (iOPB_MnGrant) OPB_state <= OPB_MASTER_SELECT;
		  OPB_MASTER_SELECT:
		    if (iOPB_xferAck) OPB_state <= OPB_READY;
		  default:
		    OPB_state <= OPB_READY;
      endcase
	 end
  end		  
  
 
  //-------------------------------------------------------------------------------------------------------
  //  REGISTERS
  //-------------------------------------------------------------------------------------------------------
  
  parameter NUMBER_OF_REGISTERS = 24;
  parameter STATUS_REGISTER = 21;
  parameter FRAME_REGISTER = 23;
  
  reg [7:0] registers [NUMBER_OF_REGISTERS - 1:0];
  reg registers_w [NUMBER_OF_REGISTERS - 1:0];
  
  wire [4:0] register_index;

  assign register_index = (iOPB_ABus - C_BASEADDR) >> 2;
  
  always @(posedge iOPB_clk)
  begin
    if (OPB_state == OPB_RESET_0)
	 begin
      registers[0] <= 8'h10;
		registers[1] <= 8'h00;
		registers[2] <= 8'h00;
		registers[3] <= 8'h00;
		registers[4] <= 8'h10;
		registers[5] <= 8'h00;
		registers[6] <= 8'h00;
		registers[7] <= 8'h00;
		registers[8] <= 8'h10;
		registers[9] <= 8'h00;
		registers[10] <= 8'h00;
		registers[11] <= 8'h00;
		registers[12] <= 8'h10;
		registers[13] <= 8'h00;
		registers[14] <= 8'h00;
		registers[15] <= 8'h00;
		registers[16] <= 8'h00;
		registers[17] <= 8'h00;
		registers[18] <= 8'h00;
		registers[19] <= 8'h00;
		registers[20] <= 8'h00;
		registers[21] <= 8'h0f;
		registers[22] <= 8'h00;
		registers[23] <= 8'h00;
	end else
	 if ( (OPB_state == OPB_READY) && (iOPB_select & ValidAddr) && (!iOPB_RNW) )
    begin
      registers[register_index] <= iOPB_DBus[7:0]; 
	 end
   end

  integer index;		
		
  always @*
  begin
	 for (index = 0; index < NUMBER_OF_REGISTERS; index = index + 1)
	 begin
	   registers_w[index] = 0;
	 end
	 if (OPB_state == OPB_SLAVE_WRITE)
	   registers_w[register_index] = 1;
	 else registers_w[register_index] = 0;
  end
	  
   //-------------------------------------------------------------------------------------------------------
  //  OPB SIGNALS
  //-------------------------------------------------------------------------------------------------------
   
  wire OPB_ack; 
  
  assign OPB_ack = (OPB_state == OPB_MASTER_SELECT) && (iOPB_xferAck);
 
  assign oSln_DBus = (OPB_state == OPB_SLAVE_READ) ? { {(OPB_WIDTH - 8) {1'b0}}, registers[register_index]}:0;
  assign oSln_xferAck = (OPB_state == OPB_SLAVE_READ || OPB_state == OPB_SLAVE_WRITE);
  
  assign oMn_request = (OPB_state == OPB_MASTER_REQUEST);
  assign oMn_ABus = (OPB_state == OPB_MASTER_SELECT) ? DMA_addr:0;  
  assign oMn_DBus = 0;
  assign oMn_Select = (OPB_state == OPB_MASTER_SELECT) ? 1:0;
  assign oMn_DBusEn = 0;
  assign oMn_BE = (OPB_state == OPB_MASTER_SELECT) ? 4'b1111:0;
  assign oMn_RNW = (OPB_state == OPB_MASTER_SELECT) ? 1:0;  

 
  //-------------------------------------------------------------------------------------------------------
  //  DMA SIGNALS Synchronization: Crossing NES CPU - OPB Clock domains. Each DMA signal is synchronous 
  //                               has corresponding synhronous instance in each domain
  //-------------------------------------------------------------------------------------------------------

  //iNES_clk driven signals
  wire NES_DMA_ack;
  wire NES_DMA_req;
    
  //iOPB_clk driven signals
  wire OPB_DMA_req_tick;
  reg OPB_ack_hold; 

  //NES to OPB synchronizer for DMA request
  sampler nes_opb(
    .iClk(iOPB_clk),
	 .iSignal(NES_DMA_req),
	 .oSignal(OPB_DMA_req_tick)
  );
  
  always @(posedge iOPB_clk or posedge iOPB_rst)
  begin
    if (iOPB_rst) OPB_DMA_req <= 0;
	 else begin
	   if (OPB_ack) OPB_DMA_req <= 0;
	   else if (OPB_DMA_req_tick) OPB_DMA_req <= 1;
	 end
  end

  //OPB to NES synchronizer for OPB acknowledge
  sampler opb_nes(
    .iClk(iNES_clk),
	 .iSignal(OPB_ack_hold),
	 .oSignal(NES_DMA_ack)
  );
  
  always @(posedge iOPB_clk or posedge iOPB_rst)
  begin
    if (iOPB_rst) OPB_ack_hold <= 0;
	 else begin
      if (NES_DMA_ack) OPB_ack_hold <= 0;
	   else if (OPB_ack) OPB_ack_hold <= 1;
	 end
  end
  
  //-------------------------------------------------------------------------------------------------------
  //  DMA DATA and ADDRESS Transformation
  //-------------------------------------------------------------------------------------------------------
  
  wire [15:0] NES_DMA_addr;
  reg [7:0] OPB_DMA_data;
  
  wire [15:0] NES_DMA_addr_relative;
  wire [1:0] DMA_BE;
  reg [7:0] OPB_data_byte;
  
  assign NES_DMA_addr_relative = NES_DMA_addr - 16'h8000;
  assign DMA_BE = NES_DMA_addr_relative[1:0];
  assign DMA_addr = DMC_MEMORY + {(NES_DMA_addr_relative >> 2), 2'b00};  
  
  always @*
  begin
    case (DMA_BE)
	   0 : OPB_data_byte = iOPB_DBus[31:24];
		1 : OPB_data_byte = iOPB_DBus[23:16];
	   2 : OPB_data_byte = iOPB_DBus[15:8];
      3 : OPB_data_byte = iOPB_DBus[7:0];
	 endcase  
  end
  
  always @(posedge iOPB_clk)
  begin
    if (OPB_ack) 
	 begin
	   OPB_DMA_data <= OPB_data_byte;
	 end
  end

  //-------------------------------------------------------------------------------------------------------
  //  APU
  //-------------------------------------------------------------------------------------------------------
 
  wire [15:0] NES_apu_out;
  wire NES_IRQ;
 
  apu apu(
    .iClk(iNES_clk),
	 .iReset(reset_channels),
	 .iRegister4000(registers[0]),
	 .iRegister4001(registers[1]),
	 .iRegister4002(registers[2]),
	 .iRegister4003(registers[3]),
	 .iRegister4004(registers[4]),
	 .iRegister4005(registers[5]),
	 .iRegister4006(registers[6]),
	 .iRegister4007(registers[7]),
	 .iRegister4008(registers[8]),
	 .iRegister400A(registers[10]),
    .iRegister400B(registers[11]),
	 .iRegister400C(registers[12]),
	 .iRegister400E(registers[14]),
	 .iRegister400F(registers[15]),
	 .iRegister4010(registers[16]),
	 .iRegister4011(registers[17]),
	 .iRegister4012(registers[18]),
	 .iRegister4013(registers[19]),
	 .iRegister4015(registers[21]),
	 .iRegister4017(registers[23]),
	 .iW4000(registers_w[0]),
	 .iW4001(registers_w[1]),
	 .iW4002(registers_w[2]),
	 .iW4003(registers_w[3]),
	 .iW4004(registers_w[4]),
	 .iW4005(registers_w[5]),
	 .iW4006(registers_w[6]),
	 .iW4007(registers_w[7]),
	 .iW4008(registers_w[8]),
	 .iW400A(registers_w[10]),
	 .iW400B(registers_w[11]),
	 .iW400C(registers_w[12]),
	 .iW400E(registers_w[14]),
	 .iW400F(registers_w[15]),
	 .iW4010(registers_w[16]),
	 .iW4011(registers_w[17]),
	 .iW4012(registers_w[18]),
	 .iW4013(registers_w[19]),
	 .iW4015(registers_w[21]),
	 .iW4017(registers_w[23]),
	 .iDMA_ack(NES_DMA_ack),
	 .iDMA_data(OPB_DMA_data),
	 .oDMA_req(NES_DMA_req),
	 .oDMA_addr(NES_DMA_addr),
	 .oData(NES_apu_out),
	 .oIRQ(NES_IRQ)
  );
  
  //-------------------------------------------------------------------------------------------------------
  //  AC 97 CONTROLLER
  //-------------------------------------------------------------------------------------------------------

  ac97_core ac97(
    .Reset(iOPB_rst),
	 `ifdef SIMULATION
	 .AC97_Bit_Clk(iOPB_clk),
	 `else
	 .AC97_Bit_Clk(iAC97_Bit_Clk),
	 `endif
	 .AC97_Sync(oAC97_Sync),
	 .AC97_SData_Out(oAC97_SData_Out),
	 .AC97_SData_In(iAC97_SData_In),
	 // AC97 register interface
	 .AC97_Reg_Addr(AC97_Reg_Addr),
	 .AC97_Reg_Write_Data(16'b0),
	 .AC97_Reg_Read_Data(),
	 .AC97_Reg_Read_Strobe(1'b0),
	 .AC97_Reg_Write_Strobe(AC97_Reg_Write_Strobe),
	 `ifdef SIMULATION
	 .AC97_Reg_Busy(),
	 `else
	 .AC97_Reg_Busy(AC97_Reg_Busy),
	 `endif
	 .AC97_Reg_Error(),
	 .AC97_Reg_Read_Data_Valid(),
	 //Playback signal interface
	 .PCM_Playback_Left(NES_apu_out),
	 .PCM_Playback_Right(NES_apu_out),
	 .PCM_Playback_Left_Valid(1'b1),
	 .PCM_Playback_Right_Valid(1'b1),
	 .PCM_Playback_Left_Accept(),
	 .PCM_Playback_Right_Accept(),
	 //Record signal interface
	 .PCM_Record_Left(),
	 .PCM_Record_Right(),
	 .PCM_Record_Left_Valid(),
	 .PCM_Record_Right_Valid(),
	 //DEBUG
	 .DEBUG(),
	 //rdy
	 `ifdef SIMULATION
	 .CODEC_RDY()
	 `else
	 .CODEC_RDY(AC97_codec_ready)
	 `endif
  );

  `ifdef SIMULATION
  assign AC97_codec_ready = 1;
  assign AC97_Reg_Busy = 0;
  `endif        	
	    	
endmodule


